package ClientRegistrar;

sub new
{
    my($class) = @_;
    my($self);

    #
    # handler_clients maps from handler name to client list
    # client_handlers maps from client gp to handler name list
    #
    $self = { handler_clients => {},
	      client_handlers => {} };

    bless($self, $class);


    return $self;
}

sub register_client
{
    my($self, $client_id, $client_ref, @event_list) = @_;
    my($event);

    if ($debug_level >= 2) {
	$str = "$client_ref $client_id @event_list\n";
	print "Registrar: Got register_client $str";
    }

    for $event (@event_list) {
	push(@{$self->{handler_clients}->{$event}}, [$client_id, $client_ref]);
    }

    $self->{client_handlers}->{$client_id} = [@event_list];
}

sub unregister_client
{
    my($self, $client_id) = @_;
    my($event,  @event_list, $er, @k);

    if($::debug_level >= 2) {
	print "Registrar: Got unregister_client $client_id\n";
    }
    
    @event_list = @{$self->{client_handlers}->{$client_id}};

    if (!@event_list) {
	warn "unregister: didn't find any events for $client_id\n";
    }

    for $event (@event_list)  {
	my(@nlist, $elt);

	for $elt (@{$self->{handler_clients}->{$event}}) {
	    push(@nlist, $elt) if $elt->[0] ne $client_id;
	}

	@{$self->{handler_clients}->{$event}} = @nlist;
    }

    delete $self->{client_handlers}->{$client_id};
}

sub event
{
    my($self, $event, $data) = @_;
    my($client);

    print "Event: $event $data\n" if $::debug_level > 3;

    for $client (@{$self->{handler_clients}->{$event}})   {
	print "Invoke $event on @$client\n" if $::debug_level > 3;
	Nexus::rpc_async($client->[1], $event, $event, $data);
    }
}

1;
