#
# This is the standard LS broker object.
#

package LSBroker;
require 5.001;

use AddINC qw (/mmlarch/lib/nperl-1.06 ./lib .);

require Nexus;
require Sys::Hostname;
require POSIX;
require Sock;
require EventDriver;
require ClientRegistrar;
require Utils;

#
# Globals
#


$nextfh = $POSIX::gensym;
$myconfigdir = $ENV{'HOME'}."/.labspace/";
$myconfigfile = $myconfigdir.join('_', split(' ', $my_name));


$dval = 0;
$Nexus::debug = $dval;
$Nexus::RRT::debug = $dval;
$Nexus::DeepPacker::debug = $dval;
$Nexus::RemoteReference::debug = $dval;

Nexus::initialize();

%LSBHandlers = (
		'register_module', \&register_module,
		'unregister_module', \&unregister_module,
		'start_module', \&start_module,
		'stop_module', \&stop_module
		);

%LSBSpecial = (		     
	       'register_module_c', \&register_module_c,
	       'unregister_module_c', \&unregister_module_c
	       );

#
# Set up the event driver
#

sub new {
    my($class, $name, $nexus_port, $handlerlist, $speciallist) = @_;

    my($tfp) = $myconfigdir.join('_', split(' ', $name));

    my($ED) = new EventDriver;

    my($self) = {
	name => $name,
	broker_url => 0,
	port => $nexus_port,
	configfn => $tfp,
	id => Sys::Hostname::hostname()."/$$",
	ed => $ED,
	attach_func => \&attach_func,
	modules => {},
	handlerlist => {%LSBHandlers, %{$handlerlist}},
	speciallist => {%LSBSpecial,  %{$speciallist}}
    };

    $ED->register_nexus();

    bless $self, $class;
}

sub initialize
{
    my($self, @list_to_start) = @_;

    while (($handler, $func) = each %{$self->{handlerlist}}) {
      Nexus::register_handler($handler, $func);
    }

    while (($handler, $func) = each %{$self->{speciallist}}) {
      Nexus::register_handler($handler, $func, unpack_buffer => 0);
    }

    {
	my($rc, $port, $host) = Nexus::allow_attach($self->{port}, 
						    sub { $self->attach_func(@_); });
if ($rc != 0) {
    ($rc, $port, $host) = Nexus::allow_attach(0, 
					      sub { $self->attach_func(@_); } );
if ($rc != 0) {
    die "Nexus::allow_attach failed with rc=$rc\n";
}
}
$self->{port} = $port;
$self->{broker_url} = "x-nexus://$host:$self->{port}/";
print "URL for the Broker: $self->{broker_url}\n" if ($::debug_level);
}

$self->dump_config();    

for $module (@list_to_start) {
    &start_module($self, $module);
}
}

sub run {
    my($self) = @_;
    my($ED) = $self->{ed};

    while (1) {
	$ED->poll();
	print "nexus poll\n" if ($::debug_level >= 4);
    }    
}

#
# Local routines that can be called remotely
#

sub register_module
{
    my($lr, $gp, $name) = @_;
    
    print "Got register_module handler for $name gp=$gp\n";

    $lr->{modules}->{perl}->{$name} = $gp;
    
    &spam_modules_default($lr);
    &spam_modules_c($lr);

    return(undef);
}

sub unregister_module
{
    my($lr, $r_id, $r_name) = @_;
    
    print "Got unregister_module for $r_name gp=$r_id\n";

    delete($lr->{modules}->{perl}->{$r_name});

    &spam_modules_default($lr);
    &spam_modules_c($lr);

    return(undef);
}

sub register_module_c
{
    my($lr, $buffer) = @_;
    my($len, $name, $rgp);

    $buffer->get_global_pointer(\$rgp);
    $buffer->get_int(\$len);
    $buffer->get_string(\$name, $len);

    if($::debug_level >= 3) {
	print "Got register_module_c handler for $name gp=$rgp\n";
    }

    $lr->{modules}->{c}->{$name} = 
	new Nexus::RemoteReference($rgp, "CModule");

    &spam_modules_default($lr);
    &spam_modules_c($lr);

    print "returning from register module c\n";

    return(undef);
}

sub unregister_module_c
{
    my($lr, $buffer) = @_;
    my($len, $name, $rgp);

    $buffer->get_global_pointer(\$rgp);
    $buffer->get_int(\$len);
    $buffer->get_string(\$name, $len);

    if ($::debug_level >= 3) {
	print "Got unregister_module_c for $name gp=$rgp\n";
    }
    delete($lr->{modules}->{c}->{$name});

    &spam_modules_default($lr);
    &spam_modules_c($lr);

    return(undef);
}

sub start_module {
    my($lr, $name, @args) = @_;


    if(!fork()) {
	print "Starting $name, with @args\n" if $::debug_level;
	exec($name, @args);
    }
    return(undef);
}

sub stop_module {
    my($lr, $name) = @_;

    print "Stopping modules isn't implemented yet...";
    print "you'll have to crash $name\n";

    return(undef);
}

#
# Local routine to call remote functions
#

sub spam_modules_default
{
    my($self) = @_;
    my($nme, $ok);
    my($lts) = {%{$self->{modules}->{perl}}, %{$self->{modules}->{c}}};

    for $nme (keys %{$self->{modules}->{perl}}) {
 	if ($::debug_level >= 3) {
 	    print "Name: ", $nme, "\nGP: ", 
	    $self->{modules}->{perl}->{$nme}, 
 	    " update_modules\n";
	}
	$ok = Nexus::rpc($self->{modules}->{perl}->{$nme}, 
			 'update_modules', %{$lts});
    }
}

sub spam_modules_c
{
    my($self) = @_;
    my($size, $n_elts, $s, $e, $nme, $len);
	    
    @foo = keys %{$self->{modules}->{c}};
    @notfoo = keys %{$self->{modules}->{perl}};
    $num_c = @foo;
    $num_perl = @notfoo;
    $num_gps = $num_c + $num_perl;

    for $nme (@foo) {
	my($dgp) = $self->{modules}->{c}->{$nme};
	$size = $n_elts = 0;
	
 	if ($::debug_level >= 3) {
 	    print "Name: ", $nme, "\nGP: ", $dgp, " update_modules\n";
	}

	my($buffer) = Nexus::init_rsr($dgp, 'update_modules');

	($s, $e) = $buffer->sizeof_int(1);
	$size += $s; $n_elts += $e;

	for $lang (keys %{$self->{modules}}) {
	    print "Lang $lang\n";
	    for $en (keys %{$self->{modules}->{$lang}}) {
		my($sgp) = $self->{modules}->{$lang}->{$en};

		($s, $e) = $buffer->sizeof_int(1);
		$size += $s; $n_elts += $e;
		($s, $e) = $buffer->sizeof_char(length($en));
		$size += $s; $n_elts += $e;
		$hgp = $sgp->gp();
	print "VAL $sgp is GP $hgp\n";
		($s, $e) = $buffer->sizeof_global_pointer($hgp);
		$size += $s; $n_elts += $e;
	    }
	}	

	$buffer->set_size($size, $n_elts);
	$buffer->put_int($num_gps);

	for $lang (keys %{$self->{modules}}) {
	    for $en (keys %{$self->{modules}->{$lang}}) {
		my($sgp) = $self->{modules}->{$lang}->{$en};
		
		$buffer->put_int(length($en));
		$buffer->put_string($en);

		$hgp = $sgp->gp();
		$buffer->put_global_pointer($hgp);
	    }
	}
	$buffer->send_rsr();
    }
}

sub dump_config
{
    my($self) = @_;
    my($fd) = \POSIX::gensym();

    open($fd, ">$self->{configfn}") || die "Couldn't open config file: !$\n";
    print $fd "Name: $self->{name}\n";
    print $fd "Broker: $self->{broker_url}\n";
    close($fd);
}

sub attach_func
{
    my($self, $url, $gpref) = @_;

    print "Calling attach_func\n" if $debug_level;
    $$gpref = $self;
    return 0;
}
