#
# This is the standard LS broker object.
#

package LSBroker;
require 5.001;

use AddINC qw (/mmlarch/lib/nperl-1.06 ./lib .);

require Nexus;
require Sys::Hostname;
require POSIX;
require Sock;
require EventDriver;
require ClientRegistrar;
require Utils;

#
# Globals
#


$nextfh = $POSIX::gensym;
$myconfigdir = $ENV{'HOME'}."/.labspace/";
$myconfigfile = $myconfigdir.join('_', split(' ', $my_name));


$dval = 0;
$Nexus::debug = $dval;
$Nexus::RRT::debug = $dval;
$Nexus::DeepPacker::debug = $dval;
$Nexus::RemoteReference::debug = $dval;

Nexus::initialize();

%LSBHandlers = (
		'register_module', \&register_module,
		'unregister_module', \&unregister_module,
		'start_module', \&start_module,
		'stop_module', \&stop_module
		);

#
# Set up the event driver
#

sub new {
    my($class, $name, $nexus_port, $handlerlist) = @_;

    my($tfp) = $myconfigdir.join('_', split(' ', $name));

    my($ED) = new EventDriver;

    my($self) = {
	name => $name,
	broker_url => 0,
	port => $nexus_port,
	configfn => $tfp,
	id => Sys::Hostname::hostname()."/$$",
	ed => $ED,
	attach_func => \&attach_func,
	modules => {},
	handlerlist => {%LSBHandlers, %{$handlerlist}},
    };

    $ED->register_nexus();

    bless $self, $class;
}

sub initialize
  {
    my($self, @list_to_start) = @_;
    my($ED) = $self->{ed};

    
    while (($handler, $func) = each %{$self->{handlerlist}}) {
      Nexus::register_handler($handler, $func);
    }
    
    if (1) {
      my($rc, $port, $host) = Nexus::allow_attach($self->{port}, 
						  sub { $self->attach_func(@_) });
      if ($rc != 0) {
	($rc, $port, $host) = Nexus::allow_attach(0, 
						  sub { $self->attach_func(@_); } );
	if ($rc != 0) {
	  die "Nexus::allow_attach failed with rc=$rc\n";
	}
      }
      $self->{port} = $port;
      $self->{broker_url} = "x-nexus://$host:$self->{port}/";
      print "URL for the Broker: $self->{broker_url}\n" if ($::debug_level);
    }
    
    $self->dump_config();    
    
    for $module (@list_to_start) {
      $num = keys %{$self->{modules}};
      &start_module($self, $module);
      while($num == keys %{$self->{modules}})
	{    
	  $ED->poll();
	}
    }
  }

sub run {
    my($self) = @_;
    my($ED) = $self->{ed};

    while (1) {
	$ED->poll();
	print "nexus poll\n" if ($::debug_level >= 4);
    }    
}

#
# Local routines that can be called remotely
#

sub register_module
{
    my($lr, $gp, $name) = @_;
    
    print "Got register_module handler for $name gp=$gp\n";

    $lr->{modules}->{$name} = $gp;
    
    &spam_modules($lr);

    return(undef);
}

sub unregister_module
{
    my($lr, $r_id, $r_name) = @_;
    
    print "Got unregister_module for $r_name gp=$r_id\n";

    delete($lr->{modules}->{$r_name});

    &spam_modules($lr);

    return(undef);
}

sub start_module {
    my($lr, $name, @args) = @_;


    if(!fork()) {
	print "Starting $name...\n" if $::debug_level;
	exec($name);
	die "Exec $name Failed: $!\n";
    }
    return(undef);
}

sub stop_module {
    my($lr, $name) = @_;

    print "Stopping modules isn't implemented yet...";
    print "you'll have to crash $name\n";

    return(undef);
}

#
# Local routine to call remote functions
#

sub spam_modules
{
    my($self) = @_;
    my($nme, $ok);

    for $nme (keys %{$self->{modules}}) {
 	if ($::debug_level >= 3) {
 	    print "Name: ", $nme, "\nGP: ", $self->{modules}->{$nme}, 
 	    " update_modules ", %{$self->{modules}}, "\n";
	}
	Nexus::rpc_async($self->{modules}->{$nme}, 
			 'update_modules', %{$self->{modules}});
    }
}

sub dump_config
{
    my($self) = @_;
    my($fd) = \POSIX::gensym();

    open($fd, ">$self->{configfn}") || die "Couldn't open config file: !$\n";
    print $fd "Name: $self->{name}\n";
    print $fd "Broker: $self->{broker_url}\n";
    close($fd);
}

sub attach_func
{
    my($self, $url, $gpref) = @_;

    print "Calling attach_func\n" if $debug_level;
    $$gpref = $self;
    return 0;
}
