
#
# This is a standard module LS
# 

#
#
#  $LSModule = {
#		$me_name,
#		$me_handlers,
#		$client_termination_function,
#		%Modules, 
#		...
#		};

package LSModule;
require 5.001;
use AddINC qw (/mmlarch/lib/nperl-1.06 /usr/local/lib/Tk ./lib .);
require Nexus;
require Sys::Hostname;
require POSIX;

#
# Nexus stuff
#

$dval = 0;
$Nexus::debug = $dval;
$Nexus::RRT::debug = $dval;
$Nexus::DeepPacker::debug = $dval;
$Nexus::RemoteReference::debug = $dval;

Nexus::initialize();

%LSMHandlers = (
		'update_modules', \&update_modules,
		);
$myconfigdir = $ENV{'HOME'}."/.labspace/";

sub new
{
    my($class, $name, $broker_name, $ed, $handlerlist, $termfunc) = @_;

    my($bcfg) = \POSIX::gensym();
    my($bcfgfn) = join('_', split(' ', $broker_name));

    open($bcfg, $myconfigdir.$bcfgfn) 
	|| die "Couldn't open broker config.\n";
    @bconfig = <$bcfg>;
    my(@matchs) = grep(/Broker:/, @bconfig);
    for $i (@matchs) {
        ($label, $burl) = split(' ', $i);
    }
    close($bcfg);

    my($tfp) = $myconfigdir.join('_', split(' ', $name));
	
    my($self) = {
	broker_url => $burl,
	broker => 0,
	name => $name,
	configfn => $tfp,
	id => Sys::Hostname::hostname()."/$$",
	termfunc => $termfunc,
	modules => {},
	ed => $ed,
	task_list => {},
	handlerlist => {%LSMHandlers, %{$handlerlist}},
	termwfd => \POSIX::gensym(),
	termrfd => \POSIX::gensym(),
    };

    $ed->register_nexus();

    bless $self, $class;
}

sub initialize
{
    my($self) = @_;
    my($event, $handler, $ok);
    my($fd) = \POSIX::gensym();

    $self->setup_signals();
    
    while(($event, $handler) = each %{$self->{handlerlist}}) {
	print "Handler $handler for Event $event\n" if ($::debug_level >= 3);
      Nexus::register_handler($event, $handler);
    }

    print "Attaching to $self->{broker_url}\n" if ($::debug_level >= 3);
    $self->{broker} = Nexus::attach($self->{broker_url});

    if (!$self->{broker}) {
	warn "Attach failed\n";
      Nexus::exit(0,1);
    }

    open($fd, ">$self->{configfn}") 
	|| die "Couldn't open config file: $self->{configfn}\n";
  
    print $fd "Name: $self->{name}\n";	
    print $fd "Broker: $self->{broker_url}\n";	

    close($fd);

#
# We assume these functions on these other objects, this should be a step 
#   down into a tighter protocol loop between the module and the broker.
#
    Nexus::rpc_async($self->{broker}, 'register_module', $self, $self->{name});

    return 1;
}

sub setup_signals
{
    my($self) = shift;
    my($fd) = $self->{termwfd};
    
    $SIG{'INT'} =  sub {print $fd "Hey there!\n"; };
    $SIG{'TERM'} = sub {print $fd "Hey there!\n"; };
    $SIG{'HUP'} = sub {print $fd "Hey there!\n"; };

    pipe($self->{termrfd}, $self->{termwfd});

    select((select($self->{termwfd}), $| = 1)[0]);

    $self->{ed}->register($self->{termrfd}, sub { $self->terminate(); } );
}

sub terminate
{
    my($self) = shift;
    my($p);

    $ok = Nexus::rpc($self->{modules}->{"Elab IO Gateway"}, 
		     "unregister_client", $self->{id});
    $ok = Nexus::rpc($self->{broker}, "unregister_module", $self->{id}, 
		     $self->{name});
    for $p (keys %{$self->{modules}}) {  delete($self->{modules}->{$p});  }
    undef $self->{broker};
    &{$self->{termfunc}}();
    Nexus::exit();
}

sub update_modules
{
    my($self, %modulelist) = @_;
    
    $self->{modules} = \%modulelist;
    if($::debug_level >= 3) {
	while(($key, $val) = each %modulelist) {
	    print "Key = $key Value = $val\n";
	}
    }
    print "Updating Modules\n" if ($::debug_level >= 3);
    return(undef);
}

sub dump_nested
{
    my($self, $ref, $indent) = @_;
    my $spaces = ' ' x ($indent * 2);

    if (ref($ref) eq "ARRAY")
    {
	my($l,$r);

	$l = @$ref;
	print "${spaces}Array[$l]:\n" if ($debug_level);
	for $r (@$ref)
	{
	    &dump_nested($r, $indent + 1);
	}
    }
    elsif (ref($ref) eq "HASH")
    {
	my($k, $v);

	print "${spaces}Hash:\n" if ($debug_level);
	while (($k, $v) = each(%$ref))
	{
	    print "${spaces}Key=\"$k\"\n" if ($debug_level);
	    &dump_nested($v, $indent + 1);
	}
    }
    else
    {
	print "${spaces}Other: $ref\n" if ($debug_level);
    }
}

sub grok_alist
{
    my($self, $alist) = @_;
    my($aelt, $hash);

    return undef if ref($alist) ne 'ARRAY';

    $hash = {};

    for $aelt (@$alist)
    {
	return undef if ref($aelt) ne 'ARRAY';
	return undef if @$aelt != 2;
	$hash->{$aelt->[0]} = $aelt->[1];
    }
    return $hash;
}

sub run
{
    my($self) = shift;
    my($ED) = $self->{ed};

    while (1) {
	$ED->poll();
	print "nexus poll\n" if ($debug_level >= 4);
    }
    
    $self->terminate();
}    

1;
