
package MudConn;
require Sock;
#
# Encapsulation for a mud connection.
#
# We include a state machine to keep things straight.
# It's pretty simple right now.
#
# 
#

#
# States:
#
my $Unconnected = 1;
my $Connected = 2;

my %MethodTable = (#connect => "connect",
		   update_interval => "set_update_interval",
		   server => "server");

my $NextID = "MG12345";

sub new
{
    my($class, $ed) = @_;
    my($self);

    $self = {};
    bless $self, $class;

    $self->{fh} = undef;
    $self->{update_interval} = 30 * 1000;

    $self->{state} = $Unconnected;

    $self->{ed} = $ed;

    $self->{id} = $NextID++;

    return $self;
}

sub id
{
    my($self) = @_;
    return $self->{id};
}

sub process_commands
{
    my($self, $cmds, $headers, $header_text, $body_text) = @_;
    my($cmd, $addr, $subject);
    
    $addr = $$headers{"reply-to"};
    if (!$addr)
    {
	$addr = $$headers{"from"};
    }
    if (!$addr)
    {
	die "cannot find return address!\n";
    }

    $addr = &rfc822::first_addr_spec($addr);

    $subject = $$headers{"subject"};

    $self->{addr} = $addr;

#    print "process_commands:\n";

    for $cmd (@$cmds)
    {
#	print "    $cmd\n";
	($cmd, @args) = split(/\s+/, $cmd);

	if (defined($method = $MethodTable{$cmd}))
	{
	    $self->$method(@args);
	}
	else
	{
	    $self->verbatim($cmd, @args);
	}
    }
}

sub server
{
    my($self, $server, $port) = @_;
    my($fh);

    if ($self->{state} != $Unconnected)
    {
	$self->error($addr, $subject, "Already connected");
	return;
    }

    $fh = new Sock();
    $fh->connect($server, $port);

    $self->{state} = $Connected;
    
    $self->{buffer} = '';

    $self->{fh} = $fh;
    $self->{ed}->register($fh, \&input_handler, $self);
    $self->{ed}->register_timer(1000, \&timer_handler, $self);
}

sub set_update_interval
{
    my($self, $interval) = @_;
    if ($interval =~ /^\d+$/)
    {
	$interval *= 1000;
	$self->{update_interval} = $interval;
	$self->{buffer} .= "[[Set update interval to $interval]]\n";
    }
    else
    {
	print "Invalid update command $interval\n";
    }
}

sub verbatim
{
    my($self, @args) = @_;
    my($fh);

#    print "verbatim state=", $self->{state}, "\n";
    if ($self->{state} != $Connected)
    {
	$self->error($addr, $subject, "Not connected");
	return;
    }

    $fh = $self->{fh};
#    print "Got fh $fh\n";
    print $fh join(" ", @args), "\n";
}

sub error
{
    my($self, @args) = @_;

    print "Error: @args\n";
}

sub timer_handler
{
    my($obj) = @_;
    $obj->timer();
}

sub input_handler
{
    my($obj, $fh) = @_;
    $obj->input($fh);
}

sub input
{
    my($self, $fh) = @_;
    my($buf, $n);
    
#    print "Got input!\n";
    $n = sysread($fh, $buf, 4096);
#    print "read n=$n $buf\n";
    $self->{buffer} .= $buf;
}

sub timer
{
    my($self) = @_;

#    print "Timer!\n";
#    print "buffer is\n";
#    print $self->{buffer};

    if ($self->{buffer} ne '')
    {
	print "Sending mail\n";
	&main::smail($self->{addr}, $main::MudmailAddress,
		     "MudMail GW MudID=" . $self->id(), $self->{buffer});
    }
    else
    {
#	print "Not updating, nothing there\n";
    }

    $self->{buffer} = '';
    $self->{ed}->register_timer($self->{update_interval},
				\&timer_handler, $self);
}
