#include "JpegReceiver.h"
#include "Source.h"
#include "net.h"

static int debug = 0;

JpegReceiver::JpegReceiver(Network *net, myim *image) : NetReceiver(net)
{
  rvts_ = new u_char[16384];
  netimage = image;
  max_s = 0;
}

void JpegReceiver::receive_rtp()
{
    struct sockaddr_in from;
    int fromlen;
    int n;
    unsigned char *addrC;
    rtphdr *rh;
    Source *source;
    unsigned short flags, seqno;
    unsigned long ts, ssrc;
    int fsrc;

    fromlen = sizeof(from);
    n = recvfrom(network_->rchannel(), buf_, sizeof(buf_), 0,
		 (struct sockaddr *) &from, &fromlen);

    addrC = (unsigned char *) &from.sin_addr.s_addr;
    if(debug)
      fprintf(stderr, "RTP: %d bytes from %d.%d.%d.%d\n",
	     n, addrC[0], addrC[1], addrC[2], addrC[3]);

    rh = (rtphdr *) buf_;

    flags = ntohs(rh->rh_flags);
    seqno = ntohs(rh->rh_seqno);
    ts = ntohl(rh->rh_ts);
    ssrc = ntohl(rh->rh_ssrc);

    if ((flags & 0x007f) != RTP_PT_JPEG)
    {
	printf("Can only handle JPEG (not %s)\n", 
	       payload_types[flags & 0x007f]);
	return;
    }

    if ((fsrc = find_source(ssrc)) != -1)
    {
	source = slist[fsrc];
    }
    else
    {
	source = new Source(ssrc, this, netimage);
	slist[max_s++] = source;
    }

    source->rtp_packet(buf_, n);
}

int JpegReceiver::find_source(long src)
{
  int i;
  
  for(i = 0; i < max_s; i++) {
    if(slist[i]->ssrc() == src) {
      return i;
    }
  }
  return -1;
}

void JpegReceiver::receive_rtcp()
{
    struct sockaddr_in from;
    int fromlen;
    int n;
    unsigned char *addrC;
    long ssrc;
    Source *source;
    rtcphdr *rh;
    char *bufptr;
    int fsrc;

    fromlen = sizeof(from);
    n = recvfrom(network_->rctrl(), buf_, sizeof(buf_), 0,
		 (struct sockaddr *) &from, &fromlen);

    addrC = (unsigned char *) &from.sin_addr.s_addr;

    if(debug)
      fprintf(stderr, "RTCP: %d bytes from %d.%d.%d.%d\n",
	     n, addrC[0], addrC[1], addrC[2], addrC[3]);

    bufptr = buf_;
    while (bufptr < buf_ + n)
    {
	rh = (rtcphdr *) bufptr;
	
	ssrc = ntohl(rh->rh_ssrc);
	
    if ((fsrc = find_source(ssrc)) != -1)
    {
	source = slist[fsrc];
    }
    else
    {
	source = new Source(ssrc, this, netimage);
	slist[max_s++] = source;
    }

	source->rtcp_packet(bufptr, rh->rh_len);

	bufptr += 4 * (rh->rh_len + 1);
    }
}


