#include <stdlib.h>
#include <unistd.h>
#include <osfcn.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <sys/time.h>
#include <sys/uio.h>

#ifdef _AIX
#include <sys/select.h>
#endif

#include "NetReceiver.h"

NetReceiver::NetReceiver(Network *net) : network_(net)
{
}

void NetReceiver::run()
{
    fd_set fds, nfds;
    int n;
    
    maxbw_ = 6000000;

    bw_ = double(maxbw_) / 8.;
//    rtcp_inv_bw_ = 1. / (bw_ * RTCP_SESSION_BW_FRACTION);

//    printf("rtcp_inv_bw=%g\n", rtcp_inv_bw_);

    FD_ZERO(&fds);
    FD_SET(network_->rchannel(), &fds);
    FD_SET(network_->rctrl(), &fds);

    while (1)
    {
	nfds = fds;
	n = select(32, &nfds, (fd_set *) NULL, (fd_set *) NULL,
		   (struct timeval *) NULL);
	if (n < 0)
	{
	    if (errno == EINTR)
		continue;
	    else
	    {
		perror("select");
		exit(1);
	    }
	}
	if (n > 0 && FD_ISSET(network_->rchannel(), &nfds))
	{
	    receive_rtp();
	    n--;
	}
	if (n > 0 && FD_ISSET(network_->rctrl(), &nfds))
	{
	    receive_rtcp();
	    n--;
	}
    }
}

void NetReceiver::receive_rtp()
{
    struct sockaddr_in from;
    int fromlen;
    int n;
    unsigned char *addrC;
    rtphdr *rh;
    unsigned short flags, seqno;
    unsigned long ts, ssrc;

    fromlen = sizeof(from);
    n = recvfrom(network_->rchannel(), buf_, sizeof(buf_), 0,
		 (struct sockaddr *) &from, &fromlen);

    addrC = (unsigned char *) &from.sin_addr.s_addr;
//    printf("RTP: %d bytes from %d.%d.%d.%d\n",
//	   n, addrC[0], addrC[1], addrC[2], addrC[3]);

    rh = (rtphdr *) buf_;

    flags = ntohs(rh->rh_flags);
    seqno = ntohs(rh->rh_seqno);
    ts = ntohl(rh->rh_ts);
    ssrc = ntohl(rh->rh_ssrc);

    process_rtp(rh, (u_char *) rh + sizeof(rtphdr), n - sizeof(rtphdr));
}

void NetReceiver::receive_rtcp()
{
    struct sockaddr_in from;
    int fromlen;
    int n;
    unsigned char *addrC;
    rtcphdr *rh;
    char *bufptr;

    fromlen = sizeof(from);
    n = recvfrom(network_->rctrl(), buf_, sizeof(buf_), 0,
		 (struct sockaddr *) &from, &fromlen);

    addrC = (unsigned char *) &from.sin_addr.s_addr;

//    printf("RTCP: %d bytes from %d.%d.%d.%d\n",
//	   n, addrC[0], addrC[1], addrC[2], addrC[3]);

    bufptr = buf_;
    while (bufptr < buf_ + n)
    {
	rh = (rtcphdr *) bufptr;
	
	process_rtcp(rh, (u_char *) rh + sizeof(rtcphdr),
		     4 * (rh ->rh_len + 1));

	bufptr += 4 * (rh->rh_len + 1);
    }
}

void NetReceiver::process_rtp(rtphdr *header, u_char *data, int data_len)
{
    unsigned char *indata;

    indata = data;
    printf("rtp:  flags=%02x seq=%x ts=%x ssrc=%x data_len=%d\n",
	   ntohs(header->rh_flags),
	   ntohs(header->rh_seqno),
	   ntohl(header->rh_ts),
	   ntohl(header->rh_ssrc),
	   data_len);
}

void NetReceiver::process_rtcp(rtcphdr *header, u_char *data, int data_len)
{
    unsigned char *indata;
    int dlen;

    indata = data;
    dlen = data_len;
    
    printf("rtcp: flags=%02x len=%d ssrc=%x\n",
	   ntohs(header->rh_flags),
	   ntohl(header->rh_len),
	   ntohl(header->rh_ssrc));
}
