#include "Source.h"
#include "JpegReceiver.h"

#undef DEBUG

#undef RAWIMAGE

#ifdef RAWIMAGE
 #include <bstring.h>
 #define RGBFILE "./testim.raw"
#endif

#ifndef __cplusplus
 double my_timer();
#else
extern "C" {
 double my_timer();
}
#endif

int do_time = 0;
static int frame_count = 0;
double start, stop;

Source::Source(long ssrc, JpegReceiver *receiver, myim *image) :
ssrc_(ssrc), receiver_(receiver)
{ 
  netim = image; 
  codec = NULL; 
}

long const & key(Source * const s)
{
  return s->ssrc();
}

Source::~Source()
{
}

void Source::rtp_packet(char *packet, int len)
{
    u_char *frame;
    rtphdr *rh;
    jpeghdr *jh;
    u_char *data;
    int jlen;
    int jq, jw, jht;
    JpegDecoder::config _config;
    double t1, t2, tot;

    rh = (rtphdr *) packet;
    jh = (jpeghdr *) (packet + sizeof(rtphdr));
    data = (u_char *) jh + sizeof(jpeghdr);
    jlen = len - sizeof(rtphdr) - sizeof(jpeghdr);
    
    jq = (int)jh->q;
    jw = (int)jh->width;
    jht = (int)jh->height;
    
    frame = reassembler_.reassemble(rh, (u_char *) jh, data, jlen);
    
    if(codec == NULL) {
      JpegDecoder::defaults(_config);
      _config.width = jh->width << 3;
      _config.height = jh->height << 3;
      JpegDecoder::quantizer(_config, jh->q);
      codec = JpegDecoder::create(_config);
    }


   if (frame)
    {
      if(do_time) {
	if(!(frame_count % 100)) {
	  stop = my_timer();
	  if(frame_count)
	    fprintf(stderr, "FPS: %f (%d)\n", 100 / (stop - start), 
		frame_count);
	  start = my_timer();
	}
      }
      t1 = my_timer();
      codec->decode(frame, jlen, receiver_->rvts_, receiver_->now_);
      t2 = my_timer();
#ifdef DEBUG      
      tot = t2 - t1;
      fprintf(stderr, "Decomp: %f ", tot);
#endif
      t1 = my_timer();
      copy_yuv_frame(codec, netim); 
      t2 = my_timer();
      frame_count++;
#ifdef DEBUG
      tot += t2 - t1;
      fprintf(stderr, "\tYUV->RGB: %f\tTotal: %f\n", t2 - t1, tot);
#endif
    }
}

void Source::rtcp_packet(char *packet, int len)
{
  char *pmem;
  int size;

  pmem = packet;
  size = len;
}

void Source::copy_yuv_frame(JpegDecoder *codec, myim *im)
{
  const u_char *y_data = codec->yplane();
  const u_char *u_data = codec->uplane();
  const u_char *v_data = codec->vplane();
  
  int i = codec->width() * codec->height(), j, t;

  u_char *target;

#ifdef RAWIMAGE
  FILE *fp;
  static int didit = 0;
#endif

#define UCLIMIT(x) ((t = (x)), (t &= ~(t>>31)), (t | ~((t-256) >> 31)))

  im->sizeX = codec->width();
  im->sizeY = codec->height();

  if(!im->lock) {
      int u, v;
      
      im->lock = 1;
      im->offset = (int)im->data;
      im->size = 0;

      target = im->data;

      for(j = 0; j < i; j += 2) {
	u = (int) u_data[j/2] - 128;
	v = (int) v_data[j/2] - 128;
	
	target[3 * (j + 0) + 0] = UCLIMIT((int)y_data[j+0] + u * 29049 / 16384);
	target[3 * (j + 0) + 1] = UCLIMIT((int)y_data[j+0] - (u * 5638 + v * 11700) / 16384);
	target[3 * (j + 0) + 2] = UCLIMIT((int)y_data[j+0] + 22970 * v / 16384);
	target[3 * (j + 1) + 0] = UCLIMIT((int)y_data[j+1] + u * 29049 / 16384);
	target[3 * (j + 1) + 1] = UCLIMIT((int)y_data[j+1] - (u * 5638 + v * 11700) / 16384);
	target[3 * (j + 1) + 2] = UCLIMIT((int)y_data[j+1] + 22970 * v / 16384);
      }

#ifdef RAWIMAGE
      if(!didit) {
	if((fp = fopen(RGBFILE, "w")) != NULL) {
	  if(fwrite(im->data, 3, i, fp)) {
	    fclose(fp);
	  } else {
	    perror("Can't write data");
	  }
	} else {
	  perror("Can't open file");
	}
	didit = 1;
      }
#endif
      
      im->size = 3 * i;
      im->getting_frames = 1;
      im->lock = 0;
  }
}
  
  void Source::update_seq(unsigned short seq)
{
  unsigned short newseq;

  newseq = seq;
}
