/*
 * Copyright (c) 1991-1994 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and the Network Research Group at
 *      Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * @(#) /home/mmlab1/Software/CVSMaster/robot/stex/jpeg/net.h,v 1.1.1.1 1996/06/06 16:47:07 judson Exp (LBL)
 */

#ifndef vic_net_h
#define vic_net_h

#include "inet.h"
#include <sys/socket.h>


class Crypt;

class Network {
    public:
	Network(n_long addr, u_short port, int ttl);
	virtual ~Network();
	virtual void send(u_char* buf, int len);
	virtual void sendmsg(const msghdr& mh);
	virtual void sendctrl(u_char* buf, int len);
	virtual int recv(u_char* buf, int len, n_long& from);
	virtual int recvctrl(u_char* buf, int len, n_long& from);

	inline int rchannel() const { return (rsock_); }
	inline int schannel() const { return (ssock_); }
	inline int sctrl() const { return (cssock_); }
	inline int rctrl() const { return (crsock_); }
	inline n_long addr() const { return (addr_); }
	inline n_long interface() const { return (local_); }
	inline int port() const { return (port_); }
	inline int cport() const { return (cport_); }
	inline int ttl() const { return (ttl_); }
	virtual void reset();
	inline u_int srcid() const { return (srcid_); }
	void resolve_srcid_collision();
	static void nonblock(int fd);
	inline static void crypt(Crypt* p) { crypt_ = p; }
    protected:
	virtual void dosend(u_char* buf, int len, int fd);
	virtual int dorecv(u_char* buf, int len, n_long& from, int fd);
	n_long alloc_srcid(n_long addr) const;

	n_long addr_;
	n_long local_;
	u_short port_;
	u_short cport_;
	u_int srcid_;
	int ttl_;
	int rsock_;
	int ssock_;
	int cssock_;
	int crsock_;

	static Crypt* crypt_;
};

#endif
