/* Includes */
#include <stex.h>

nexus_global_pointer_t gp_list[NUMNAMES];
char name_list[NUMNAMES][100];
extern int wait, robot_return_value;

extern int *in_stereo, *activity, *control, *new_user, *exit_now;

/* Functions */

extern void update_pan_tilt(float azi, float elev, nexus_global_pointer_t *gp);
extern void ptto(int pp, int tp, nexus_global_pointer_t *gp);

extern void update_robot(float x, float y, float maxx, float maxy,
                  nexus_global_pointer_t *gp, char *name);
/* functions that call handlers in the other caves */
/* These are for setting the state of the other cave */

void set_control(nexus_global_pointer_t *gp, int val)
{
  nexus_buffer_t buffer;
  int fa = val;
  nexus_init_remote_service_request(&buffer, gp, "sync_control",
				    nexus_handler_hash("sync_control"));
  nexus_set_buffer_size(&buffer, nexus_sizeof_int(&buffer, 1) , 1);
  nexus_put_int(&buffer, &fa, 1);
  nexus_send_remote_service_request(&buffer);
}

void set_mode(nexus_global_pointer_t *gp, int *val)
{
  nexus_buffer_t buffer;

  nexus_init_remote_service_request(&buffer, gp, "sync_mode",
				    nexus_handler_hash("sync_mode"));
  nexus_set_buffer_size(&buffer, nexus_sizeof_int(&buffer, 1) , 1);
  nexus_put_int(&buffer, val, 1);
  nexus_send_remote_service_request(&buffer);
}

void set_activity(nexus_global_pointer_t *gp, int *val)
{
  nexus_buffer_t buffer;

  nexus_init_remote_service_request(&buffer, gp, "sync_activity",
				    nexus_handler_hash("sync_activity"));
  nexus_set_buffer_size(&buffer, nexus_sizeof_int(&buffer, 1) , 1);
  nexus_put_int(&buffer, val, 1);
  nexus_send_remote_service_request(&buffer);
}

/* Handlers for calling into this particular CAVE */
/* These are for changing state here so it reflects the state of another CAVE */

void sync_control(void *address, nexus_buffer_t *buffer)
{
  nexus_get_int(buffer, control, 1);
}

void sync_mode(void *address, nexus_buffer_t *buffer)
{
  nexus_get_int(buffer, in_stereo, 1);
}

void sync_activity(void *address, nexus_buffer_t *buffer)
{
  nexus_get_int(buffer, activity, 1);
}

void sync_exit(void *address, nexus_buffer_t *buffer)
{
  nexus_get_int(buffer, exit_now, 1);
}

void register_module(nexus_global_pointer_t *to_gp, char *name)
{
  nexus_buffer_t buffer;
  nexus_global_pointer_t gp;
  int buffer_size, gp_n_elements;
  int name_len;
  char itema = 'a', itemg = 'g', itemc = 'c';
  int count = 2;

  name_len = strlen(name);

  nexus_global_pointer(&gp, NULL);
  nexus_init_remote_service_request(&buffer, to_gp, "register_module",
                                    nexus_handler_hash("register_module"));

  buffer_size = nexus_sizeof_global_pointer(&buffer, &gp, 1, &gp_n_elements);
  buffer_size += nexus_sizeof_int(&buffer, 3);
  buffer_size += nexus_sizeof_char(&buffer, name_len + 3);
  nexus_set_buffer_size(&buffer, buffer_size, gp_n_elements + 7);
  /* Array */
  nexus_put_char(&buffer, &itema, 1);
  nexus_put_int(&buffer, &count, 1);
  /* GP */
  nexus_put_char(&buffer, &itemg, 1);
  count = 1;
  nexus_put_int(&buffer, &count, 1);
  nexus_put_global_pointer(&buffer, &gp, 1);
  /* Name */
  nexus_put_char(&buffer, &itemc, 1);
  nexus_put_int(&buffer, &name_len, 1);
  nexus_put_char(&buffer, name, name_len);
  nexus_send_remote_service_request(&buffer);
  nexus_destroy_global_pointer(&gp);
}

void unregister_module(nexus_global_pointer_t *to_gp, char *name)
{
  nexus_buffer_t buffer;
  nexus_global_pointer_t gp;
  int buffer_size, gp_n_elements;
  int name_len;
  char itema = 'a', itemg = 'g', itemc = 'c';
  int count = 2;

  name_len = strlen(name);
  
  nexus_global_pointer(&gp, NULL);
  nexus_init_remote_service_request(&buffer, to_gp, "unregister_module",
                                    nexus_handler_hash("unregister_module"));

  buffer_size = nexus_sizeof_global_pointer(&buffer, &gp, 1, &gp_n_elements);
  buffer_size += nexus_sizeof_int(&buffer, 3);
  buffer_size += nexus_sizeof_char(&buffer, name_len + 3);
  nexus_set_buffer_size(&buffer, buffer_size, gp_n_elements + 7);
  /* Array */
  nexus_put_char(&buffer, &itema, 1);
  nexus_put_int(&buffer, &count, 1);
  /* GP */
  nexus_put_char(&buffer, &itemg, 1);
  count = 1;
  nexus_put_int(&buffer, &count, 1);
  nexus_put_global_pointer(&buffer, &gp, 1);
  /* Name */
  nexus_put_char(&buffer, &itemc, 1);
  nexus_put_int(&buffer, &name_len, 1);
  nexus_put_char(&buffer, name, name_len);
  nexus_send_remote_service_request(&buffer);
}

/* Handlers required to be written with NPERL packing format communication */
/* These are for communicating with the broker... */

void update_modules(void *address, nexus_buffer_t *buffer)
{
  int num, length;
  int i, junk;
  char tag_type, name[50];
  
  for(i = 0; i < NUMNAMES; i++)
    bzero((void *)name_list[i], 100);
  
#ifdef DEBUG_NEXUS
  printf("In update_modules\n");
#endif

  nexus_get_char(buffer, &tag_type, 1);
  nexus_get_int(buffer, &num, 1);

  if(num/2 > NUMNAMES) {
    fprintf(stderr, "There are too many names coming.\n");
  }    

  for(i = 0; i < num/2; i++) {
    nexus_get_char(buffer, &tag_type, 1);
    nexus_get_int(buffer, &length, 1);

    nexus_get_char(buffer, name_list[i], length);
    nexus_get_char(buffer, &tag_type, 1);
    nexus_get_int(buffer, &junk, 1);
    
    if(tag_type == 'g') {
      nexus_get_global_pointer(buffer, &gp_list[i], 1);
    } else if(tag_type == 'r') {
      nexus_get_global_pointer(buffer, &gp_list[i], 1);
      
      nexus_get_int(buffer, &junk, 1);
      nexus_get_char(buffer, name, junk);
    }
  }
  wait = 1;

#ifdef DEBUG_NEXUS
  for(i = 0; i < num/2; i++) {
    printf("Name: %s GP: %s\n", name_list[i], 
	   nexus_global_pointer_string(&gp_list[i]));
  }
#endif

  *new_user = 1;

#ifdef DEBUG_NEXUS
  printf("Leaving update_modules\n");
#endif
}


void get_ack(void *address, nexus_buffer_t *buffer)
{
  int num;
  char tag_type;

  nexus_get_char(buffer, &tag_type, 1);
  nexus_get_int(buffer, &num, 1);
  
  nexus_get_char(buffer, &tag_type, 1);
  nexus_get_int(buffer, &num, 1);
  nexus_get_int(buffer, &robot_return_value, 1);
  fprintf(stderr, "Got ACK\n");
}
