/* Includes */
#include <stex.h>

/* Functions */
void update_pan_tilt(float azi, float elev, nexus_global_pointer_t *gp)
{
  static int pp, tp, last_pp, last_tp;
  pp = MAXPAN * (azi / 180);
  tp = MAXTILT * (elev / 90);

  if((last_pp != pp) || (last_tp != tp)) {
    ptto(pp, tp, gp);
    last_pp = pp; last_tp = tp;
  }
}

void ptto(int pp, int tp, nexus_global_pointer_t *gp)
{
  nexus_buffer_t buffer;
  int buffer_size, fa = pp, sa = tp;
  char itema = 'a';
  char itemi = 'i';
  int hm = 1;

  /* Set pan position */
  nexus_init_remote_service_request(&buffer, gp, "pan-position", 
				    nexus_handler_hash("pan-position"));
  buffer_size = nexus_sizeof_char(&buffer, 2);
  buffer_size += nexus_sizeof_int(&buffer, 3);
  nexus_set_buffer_size(&buffer, buffer_size, 5);

  nexus_put_char(&buffer, &itema, 1);
  nexus_put_int(&buffer, &hm, 1);
  nexus_put_char(&buffer, &itemi, 1);
  nexus_put_int(&buffer, &hm, 1);
  nexus_put_int(&buffer, &fa, 1);
  nexus_send_remote_service_request(&buffer);

  /* Set tilt position */
  nexus_init_remote_service_request(&buffer, gp, "tilt-position",
				    nexus_handler_hash("tilt-position"));
  buffer_size = nexus_sizeof_char(&buffer, 2);
  buffer_size += nexus_sizeof_int(&buffer, 3);
  nexus_set_buffer_size(&buffer, buffer_size, 5);

  nexus_put_char(&buffer, &itema, 1);
  nexus_put_int(&buffer, &hm, 1);
  nexus_put_char(&buffer, &itemi, 1);
  nexus_put_int(&buffer, &hm, 1);
  nexus_put_int(&buffer, &sa, 1);
  nexus_send_remote_service_request(&buffer);
}

void update_robot(float x, float y, float maxx, float maxy, 
		  nexus_global_pointer_t *gp, char *name)
{
  static float last_x, last_y;
  float z;
  float degs;
  int direction;

  direction = (y > 0) ? 0: 1;

  if((last_x != x) || (last_y != y) && 
     (x > maxx) && (y > maxy)) {
    z = sqrt(x * x + y * y);

    degs = atanf(y/x);
    degs *= 180.0;
    degs /= M_PI;

    if(degs < 0) {
      degs = abs(degs);
    } else {
      degs = (180.0 - degs);
    }                       
    if((abs(x) > maxx) || (abs(y) > maxy)) {
      fprintf(stderr, "Would have called move_robot\n");
      move_robot(z, direction, gp, name);
    }
  }
}

void turn_robot(int degs, nexus_global_pointer_t *gp, char *name)
{
  nexus_buffer_t buffer;
  int buffer_size = 0, count = 1, fa;
  char itema = 'a', itemi = 'i', itemc = 'c';
  int speed = 6, len;

  /* turn robot */
  fa = degs; len = strlen(name);
  nexus_init_remote_service_request(&buffer, gp, "robot-turn",
				    nexus_handler_hash("robot-turn"));
  buffer_size = nexus_sizeof_int(&buffer, 4);
  buffer_size += nexus_sizeof_char(&buffer, 3 + len);
  nexus_set_buffer_size(&buffer, buffer_size, 8);
  /* array ? */
  nexus_put_char(&buffer, &itema, 1);
  count = 2;
  nexus_put_int(&buffer, &count, 1);
  /* Item 1: Caller name */
  nexus_put_char(&buffer, &itemc, 1);
  nexus_put_int(&buffer, &len, 1);
  nexus_put_char(&buffer, name, len);
  /* Item 2: # Degrees */
  nexus_put_char(&buffer, &itemi, 1);
  count = 1;
  nexus_put_int(&buffer, &count, 1);
  nexus_put_int(&buffer, &fa, 1);
  nexus_send_remote_service_request(&buffer);
}

void move_robot(int z, int direction, nexus_global_pointer_t *gp, char *name)
{
  nexus_buffer_t buffer;
  int buffer_size = 0, count = 1, fa, sa;
  char itema = 'a', itemi = 'i', itemc = 'c';
  int speed = 6, len;

  /* move robot */
  fa = speed; sa = z; count = 3;
  if(direction)
    nexus_init_remote_service_request(&buffer, gp, "robot-forward",
				      nexus_handler_hash("robot-forward"));
  else
    nexus_init_remote_service_request(&buffer, gp, "robot-backward",
				      nexus_handler_hash("robot-backward"));
  buffer_size = nexus_sizeof_int(&buffer, 6);
  buffer_size += nexus_sizeof_char(&buffer, 4 + len);
  nexus_set_buffer_size(&buffer, buffer_size, 11);
  /* array ? */
  nexus_put_char(&buffer, &itema, 1);
  nexus_put_int(&buffer, &count, 1);
  count = 1;
  /* Item 1: Caller Name */
  nexus_put_char(&buffer, &itemc, 1);
  nexus_put_int(&buffer, &len, 1);
  nexus_put_char(&buffer, name, len);
  /* Item 2 : Speed */
  nexus_put_char(&buffer, &itemi, 1);
  nexus_put_int(&buffer, &count, 1);
  nexus_put_int(&buffer, &fa, 1);
  /* Item 3 : Distance */
  nexus_put_char(&buffer, &itemi, 1);
  nexus_put_int(&buffer, &count, 1);
  nexus_put_int(&buffer, &sa, 1);
  nexus_send_remote_service_request(&buffer);
}

void reset_robot(nexus_global_pointer_t *gp, char *name)
{
  nexus_buffer_t buffer;
  int buffer_size = 0, count = 1, len;
  char itema = 'a', itemc = 'c';

  /* move robot */
  len = strlen(name);
  nexus_init_remote_service_request(&buffer, gp, "robot-reset",
				      nexus_handler_hash("robot-reset"));
  buffer_size = nexus_sizeof_int(&buffer, 2);
  buffer_size += nexus_sizeof_char(&buffer, 2 + len);
  nexus_set_buffer_size(&buffer, buffer_size, 5);
  /* array ? */
  nexus_put_char(&buffer, &itema, 1);
  nexus_put_int(&buffer, &count, 1);
  count = 1;
  /* Item 1: Caller Name */
  nexus_put_char(&buffer, &itemc, 1);
  nexus_put_int(&buffer, &len, 1);
  nexus_put_char(&buffer, name, len);
  nexus_send_remote_service_request(&buffer);
}

void beep_robot(nexus_global_pointer_t *gp, char *name)
{
  nexus_buffer_t buffer;
  int buffer_size = 0, n = 1, len;
  char ch = 'c', array = 'a';

  nexus_init_remote_service_request(&buffer, gp, "robot-beep",
				    nexus_handler_hash("robot-beep"));
  len = strlen(name);
  buffer_size = nexus_sizeof_char(&buffer, len + 2);
  buffer_size += nexus_sizeof_int(&buffer, 2);

  nexus_set_buffer_size(&buffer, buffer_size, len + 4);
  
  /* Array */
  nexus_put_char(&buffer, &array, 1);
  nexus_put_int(&buffer, &n, 1);

  /* item 1: name */
  nexus_put_char(&buffer, &ch, 1);
  nexus_put_int(&buffer, &len, 1);
  nexus_put_char(&buffer, name, len);

  nexus_send_remote_service_request(&buffer);
}

void set_mono(nexus_global_pointer_t *gp, int *val)
{
  nexus_buffer_t buffer;
  int buffer_size = 0, fa = 3, sa = 4;
  int count = 1;
  char item = 'i';
  char array = 'a';
  int n = 2;

  nexus_init_remote_service_request(&buffer, gp, "svs-set-afv",
				    nexus_handler_hash("svs-set-afv"));  
  buffer_size = nexus_sizeof_int(&buffer, 5);
  buffer_size += nexus_sizeof_char(&buffer, 3);
  nexus_set_buffer_size(&buffer, buffer_size, 8);
  /* array ? */
  nexus_put_char(&buffer, &array, 1);
  nexus_put_int(&buffer, &n, 1);
  /* Item 1 */
  nexus_put_char(&buffer, &item, 1);
  nexus_put_int(&buffer, &count, 1);
  nexus_put_int(&buffer, &fa, 1);
  /* Item 2 */
  nexus_put_char(&buffer, &item, 1);
  nexus_put_int(&buffer, &count, 1);
  nexus_put_int(&buffer, &sa, 1);
  nexus_send_remote_service_request(&buffer);

  *val = 0;
}

void set_stereo(nexus_global_pointer_t *gp, int *val)
{
  nexus_buffer_t buffer;
  int buffer_size = 0, fa = 1, sa = 2;
  int count = 1;
  char item = 'i';
  char array = 'a';
  int n = 2;

  nexus_init_remote_service_request(&buffer, gp, "svs-set-afv",
				    nexus_handler_hash("svs-set-afv"));
  buffer_size = nexus_sizeof_int(&buffer, 5);
  buffer_size += nexus_sizeof_char(&buffer, 3);
  nexus_set_buffer_size(&buffer, buffer_size, 8);
  /* array ? */
  nexus_put_char(&buffer, &array, 1);
  nexus_put_int(&buffer, &n, 1);
  /* Item 1 */
  nexus_put_char(&buffer, &item, 1);
  nexus_put_int(&buffer, &count, 1);
  nexus_put_int(&buffer, &fa, 1);
  /* Item 2 */
  nexus_put_char(&buffer, &item, 1);
  nexus_put_int(&buffer, &count, 1);
  nexus_put_int(&buffer, &sa, 1);
  nexus_send_remote_service_request(&buffer);

  *val = 1;
}


void zoom_in(nexus_global_pointer_t *gp)
{
  char itema = 'a';
  int num = 0, buffer_size;
  nexus_buffer_t buffer;

  nexus_init_remote_service_request(&buffer, gp, "mpc-zoom-in",
     				      nexus_handler_hash("mpc-zoom-in"));
  buffer_size = nexus_sizeof_int(&buffer, 1);
  buffer_size += nexus_sizeof_char(&buffer, 1);
  nexus_set_buffer_size(&buffer, buffer_size, 2);
  nexus_put_char(&buffer, &itema, 1);
  nexus_put_int(&buffer, &num, 1);
  nexus_send_remote_service_request(&buffer);
  nexus_init_remote_service_request(&buffer, gp, "mpc-zoom-stop",
				    nexus_handler_hash("mpc-zoom-stop"));
  buffer_size = nexus_sizeof_int(&buffer, 1);
  buffer_size += nexus_sizeof_char(&buffer, 1);
  nexus_set_buffer_size(&buffer, buffer_size, 2);
  nexus_put_char(&buffer, &itema, 1);
  nexus_put_int(&buffer, &num, 1);
  nexus_send_remote_service_request(&buffer);
}  

void zoom_out(nexus_global_pointer_t *gp)
{
  char itema = 'a';
  int num = 0, buffer_size;
  nexus_buffer_t buffer;

  nexus_init_remote_service_request(&buffer, gp, "mpc-zoom-out",
     				      nexus_handler_hash("mpc-zoom-out"));
  buffer_size = nexus_sizeof_int(&buffer, 1);
  buffer_size += nexus_sizeof_char(&buffer, 1);
  nexus_set_buffer_size(&buffer, buffer_size, 2);
  nexus_put_char(&buffer, &itema, 1);
  nexus_put_int(&buffer, &num, 1);
  nexus_send_remote_service_request(&buffer);
  nexus_init_remote_service_request(&buffer, gp, "mpc-zoom-stop",
				    nexus_handler_hash("mpc-zoom-stop"));
  buffer_size = nexus_sizeof_int(&buffer, 1);
  buffer_size += nexus_sizeof_char(&buffer, 1);
  nexus_set_buffer_size(&buffer, buffer_size, 2);
  nexus_put_char(&buffer, &itema, 1);
  nexus_put_int(&buffer, &num, 1);
  nexus_send_remote_service_request(&buffer);
}  

void focus_far(nexus_global_pointer_t *gp)
{
  char itema = 'a';
  int num = 0, buffer_size;
  nexus_buffer_t buffer;

  nexus_init_remote_service_request(&buffer, gp, "mpc-focus-far",
     				      nexus_handler_hash("mpc-focus-far"));
  buffer_size = nexus_sizeof_int(&buffer, 1);
  buffer_size += nexus_sizeof_char(&buffer, 1);
  nexus_set_buffer_size(&buffer, buffer_size, 2);
  nexus_put_char(&buffer, &itema, 1);
  nexus_put_int(&buffer, &num, 1);
  nexus_send_remote_service_request(&buffer);
  nexus_init_remote_service_request(&buffer, gp, "mpc-focus-stop",
				    nexus_handler_hash("mpc-focus-stop"));
  buffer_size = nexus_sizeof_int(&buffer, 1);
  buffer_size += nexus_sizeof_char(&buffer, 1);
  nexus_set_buffer_size(&buffer, buffer_size, 2);
  nexus_put_char(&buffer, &itema, 1);
  nexus_put_int(&buffer, &num, 1);
  nexus_send_remote_service_request(&buffer);
}  

void focus_near(nexus_global_pointer_t *gp)
{
  char itema = 'a';
  int num = 0, buffer_size;
  nexus_buffer_t buffer;

  nexus_init_remote_service_request(&buffer, gp, "mpc-focus-near",
     				      nexus_handler_hash("mpc-focus-near"));
  buffer_size = nexus_sizeof_int(&buffer, 1);
  buffer_size += nexus_sizeof_char(&buffer, 1);
  nexus_set_buffer_size(&buffer, buffer_size, 2);
  nexus_put_char(&buffer, &itema, 1);
  nexus_put_int(&buffer, &num, 1);
  nexus_send_remote_service_request(&buffer);

  nexus_init_remote_service_request(&buffer, gp, "mpc-focus-stop",
				    nexus_handler_hash("mpc-focus-stop"));
  buffer_size = nexus_sizeof_int(&buffer, 1);
  buffer_size += nexus_sizeof_char(&buffer, 1);
  nexus_set_buffer_size(&buffer, buffer_size, 2);
  nexus_put_char(&buffer, &itema, 1);
  nexus_put_int(&buffer, &num, 1);
  nexus_send_remote_service_request(&buffer);
}  

void set_exit(nexus_global_pointer_t *gp)
{
  nexus_buffer_t buffer;
  int buffer_size;
  int die = 1;

  nexus_init_remote_service_request(&buffer, gp, "terminate",
				    nexus_handler_hash("terminate"));
  buffer_size = nexus_sizeof_int(&buffer, 1);
  nexus_set_buffer_size(&buffer, buffer_size, 1);
  nexus_put_int(&buffer, &die, 1);
  nexus_send_remote_service_request(&buffer);
}
