/* Standard include stuff */
#include <stex.h>

/* Jpeg C++ objects */
#include <stex++.h>

/* Defines for getting verbose debugging or not */
#undef DEBUG_PROCS
#undef DEBUG_PANTILT
#undef DEBUG_IMAGE
#undef DEBUG_ROBOT
#undef DEBUG_CALIB
#undef DEBUG_SVS
#undef DEBUG_NEXUS
#undef DEBUG_HANDLERS
#undef DEBUG_NAMES

#define TOTMEMSIZE  20000000
#define MEMSIZE     2000000

#undef BIG_IMG

#ifdef BIG_IMG
 #define BIG_FACTOR 2.0
#else
 #define BIG_FACTOR 1.0
#endif

#define TEXT_X_SIZE ( 512 * BIG_FACTOR )
#define TEXT_Y_SIZE 512
#define WIDTH       ( 320 * BIG_FACTOR )
#define HEIGHT      240
#define TEXTSIZE    ( TEXT_X_SIZE * TEXT_Y_SIZE * 4 )
#define SUBTEXTSIZE ( WIDTH * 256 * 3 )
#define SVAL        ( 0.5  )
#define TVAL        ( 0.625 )

#define RAD(x)      ((x * M_PI)/180.0)
#define LEFT  1
#define RIGHT 2

/* C code I'm going to use */
extern "C" {
  /* cave routines */
  void app_pre_init(int argc, char **argv);
  void app_init(void);
  void app_update(void);
  void app_display(void);
  void app_exit(void);
  void app_startup(void);

  /* graphics routines */
  static void do_jpeg_network_video(myim **image, char *hostname, int port);
  static void calibrate_joystick(float *bound, float *switch_rad);
  static void create_texture(int which);
  static void build_texture(myim *image, unsigned int textnum);
  static void build_wall(void);
  static void show_joystick(float x, float y, float rad);
  static void show_menu(void);
  static int  do_menu(float swich_rad);

  /* utility routines or nexus fixes */
  static int find_gp(char *name);
  static void sig_int(...);
  static int fault_handler(void *user_arg, int fault_code);
#ifdef DEBUG_NEXUS
  extern char *nexus_global_pointer_string(nexus_global_pointer_t *g);
#endif
}

/* Handler Table */
static nexus_handler_t handlers [] =
{
  { "update_modules", 478, NEXUS_HANDLER_TYPE_NON_THREADED,
    (nexus_handler_func_t)update_modules},
  { "sync_control", 288, NEXUS_HANDLER_TYPE_NON_THREADED,
    (nexus_handler_func_t)sync_control},
  { "sync_mode", 961, NEXUS_HANDLER_TYPE_NON_THREADED,
    (nexus_handler_func_t)sync_mode},
  { "sync_activity", 396, NEXUS_HANDLER_TYPE_NON_THREADED,
    (nexus_handler_func_t)sync_activity},
  { "robot-ack", 898, NEXUS_HANDLER_TYPE_NON_THREADED,
    (nexus_handler_func_t)get_ack},
  { "terminate", 969, NEXUS_HANDLER_TYPE_NON_THREADED,
    (nexus_handler_func_t)sync_exit},
  {(char *) NULL, 0, NEXUS_HANDLER_TYPE_NON_THREADED,
   (nexus_handler_func_t) NULL},
};


/* Globals for the application */
extern nexus_global_pointer_t gp_list[NUMNAMES];
extern char name_list[NUMNAMES][100];
int wait = 0, robot_return_value;

char *shared_space;
int *in_stereo, *activity, *master, *exit_now;
int *slave, *control, *menuing, *new_user;
int *steering_turn;

myim *jleft, *jright;
float joystick_bound[2], switch_rad;
int left_port, right_port;
unsigned int left_image, right_image;
char *source_host, *right_host, *left_host;
unsigned long *fleft_buffer, *fright_buffer;
float coords[4][2], texcoords[4][2];
int dojpegleft = 0, dojpegright = 0;

char *nexus_url;
int use_broker = 0;
nexus_global_pointer_t attach_gp;
char fdbck[150];

char *opts[] = {"Stereo", "Mono", "Turn Left", "Turn Right", 
		"Move Forward", "Move Backward", "Control"};

char *idesk       = "IDesk";
char *arpa        = "CAVE";
char *refl_name   = "Reflector Controller Gateway Left";
char *refr_name   = "Reflector Controller Gateway Right";
char *pt_name     = "Pan Tilt IO Gateway";
char *robot_name  = "Robot IO Gateway";
char *svs_name    = "Serial Video Switch IO Gateway";
char *mpc_name    = "MPC IO Gateway";

extern int errno;

static void sig_int(...)
{
  exit(0);
}

/* Main */

int main(int argc, char **argv)
{
  nexus_node_t *nodes;
  int n_nodes, rc;
  int other_index;

  CAVEConfigure(&argc, argv, NULL);

  app_pre_init(argc, argv);

  if(dojpegleft)
    do_jpeg_network_video(&jleft, left_host, left_port);
  if(dojpegright)
    do_jpeg_network_video(&jright, right_host, right_port);

  CAVEInit();

  CAVEInitApplication(app_init, 0);

  CAVEDisplay(app_display, 0);

  nexus_init(&argc, &argv, "STEX_ARGS", "nx", NULL, NULL, NULL, NULL,
	     &nodes, &n_nodes);
  nexus_start();
  nexus_enable_fault_tolerance(fault_handler, NULL);

  if(use_broker) {
    rc = nexus_attach(nexus_url, &attach_gp);
    if(rc != 0) 
      {
	printf("attach to %s filed with rc = %d\n", nexus_url, rc);
	app_exit();
	CAVEExit();
	nexus_destroy_current_context(1);
	exit(1);
      }

    register_module(&attach_gp, (*master)?idesk:arpa);

    while(!wait) { nexus_poll(); } 
  }
  
  app_startup();

  while(!CAVEgetbutton(CAVE_ESCKEY)) {
    if(*new_user) {
      if(*control) {
	other_index = find_gp((*master)?arpa:idesk);
	if(other_index > -1) {
	  set_mode(&gp_list[other_index], in_stereo);
	  set_control(&gp_list[other_index], !*control);
	  set_activity(&gp_list[other_index], activity);
	}
      }
      *new_user = 0;
    }
    app_update();
  }
  
  app_exit();

  CAVEExit();

  nexus_destroy_current_context(0);
  
  return(0);
}

static void do_jpeg_network_video(myim **image, char *host, int port)
{
  n_long addr;
  unsigned char *addrC;
  int ttl;
  int ctrlsock = 1;
  Network *net;
  JpegReceiver *receiver;
  int tval;

  addr = LookupHostAddr(host);
  ttl = 4;
  
  addrC = (unsigned char *) &addr;

#ifdef DEBUG_IMAGE
  printf("Connecting JpegReceiver to: (host) %s [%d.%d.%d.%d] (port) %d\n", 
	 host, addrC[0], addrC[1], addrC[2], addrC[3], port);
#endif

  net = new mcastChannel(addr, port, ttl, ctrlsock);

  receiver = new JpegReceiver(net, *image);

  tval = (*image)->proc_id = fork();

  if(!tval) {
    signal(SIGINT, sig_int);
    signal(SIGHUP, sig_int);
    signal(SIGKILL, sig_int);
#ifdef DEBUG_PROCS
    printf("About to exec receiver->run() one eye\n");
#endif
    receiver->run();
    exit(0);
  }
#ifdef DEBUG_PROCS
  printf("Jpeg Receiver PID: %d\n", (*image)->proc_id);
#endif
}

/* Application pre-initialization and supporting routines */
void app_pre_init(int argc, char **argv)
{
  int i;

  shared_space = (char *)CAVEUserSharedMemory(TOTMEMSIZE);

  jleft = (myim *)amalloc(sizeof(myim), shared_space);
  jleft->lock = 0;
  jleft->getting_frames = 0;
  jleft->data = (unsigned char *)amalloc(SUBTEXTSIZE, shared_space);
  
  jright = (myim *)amalloc(sizeof(myim), shared_space);
  jleft->lock = 0;
  jright->getting_frames = 0;
  jright->data = (unsigned char *)amalloc(SUBTEXTSIZE, shared_space);

  steering_turn =(int *)amalloc(sizeof(int), shared_space); *steering_turn = 0;
  master =    (int *)amalloc(sizeof(int), shared_space);      *master = 0;
  slave =     (int *)amalloc(sizeof(int), shared_space);      *slave = 0;
  new_user =  (int *)amalloc(sizeof(int), shared_space);      *new_user = 0;
  exit_now =  (int *)amalloc(sizeof(int), shared_space);      *exit_now = 0;
  control =   (int *)amalloc(sizeof(int), shared_space);      *control = 0;
  in_stereo = (int *)amalloc(sizeof(int), shared_space);      *in_stereo = 0;
  menuing =   (int *)amalloc(sizeof(int), shared_space);      *menuing = -1;
  activity =  (int *)amalloc(sizeof(int), shared_space);      *activity = 0;

  for (i = 1; i < argc; i++) {
    if (strcmp(argv[i], "-jpeg_left") == 0) {
      if (i+1 >= argc || argv[i+1][0] == '-') {
	printf("-jpeg_left (No host port).\n");
      } else {
	dojpegleft = 1;
	left_image = LEFT;
	left_host = argv[i+1];
	left_port = atoi(argv[i+2]);
	i += 2;
      }
    } else if (strcmp(argv[i], "-jpeg_right") == 0) {
      if (i+1 >= argc || argv[i+1][0] == '-') {
	printf("-net_right_host (No file name).\n");
      } else {
	dojpegright = 1;
	right_image = RIGHT;
	right_host = argv[i+1];
	right_port = atoi(argv[i+2]);
	i += 2;
      }
    } else if (strcmp(argv[i], "-me") == 0) {
      if (i+1 >= argc || argv[i+1][0] == '-') {
	printf("-me (No host name).\n");
      } else {
	source_host = argv[i+1];
	i += 1;
      }
    } else if (strcmp(argv[i], "-broker") == 0) {
      if (i+1 >= argc || argv[i+1][0] == '-') {
	printf("-broker (no URL?).\n");
      } else {
	use_broker = 1;
	nexus_url = argv[i+1];
	i += 1;
      }
    } else if (strcmp(argv[i], "-master") == 0) {
      *master = 1; *slave = 0; *control = 1;
    } else if (strcmp(argv[i], "-slave") == 0) {
      *slave = 1; *master = 0; *control = 0;
    }
  }
  coords[0][0] = -5.0;
  coords[0][1] = 0.0;
  coords[1][0] = 5.0;
  coords[1][1] = 0.0;
  coords[2][0] = 5.0;
  coords[2][1] = 10.0;
  coords[3][0] = -5.0;
  coords[3][1] = 10.0;

  texcoords[0][0] = 0.0;
  texcoords[0][1] = 0.49;
  texcoords[1][0] = 0.0;
  texcoords[1][1] = 0.0;
  texcoords[2][0] = 0.61;
  texcoords[2][1] = 0.0;
  texcoords[3][0] = 0.61;
  texcoords[3][1] = 0.49;

}

/* Application initialization and supporting routines */
void app_init(void)
{
  int j;

  fleft_buffer = (unsigned long *)malloc(TEXTSIZE);
  if(fleft_buffer < 0)
    printf("Dead mem\n");
  for(j = 0; j < TEXTSIZE/4; j++)
    fleft_buffer[j] = (unsigned long)0xff00ff00;
  
  fright_buffer = (unsigned long *)malloc(TEXTSIZE);
  if(fright_buffer < 0)
    printf("Dead mem\n");
  for(j = 0; j < TEXTSIZE/4; j++)
    fright_buffer[j] = (unsigned long)0xff00ff00;
  
  if(dojpegleft)
    create_texture(LEFT);

  if(dojpegright)
    create_texture(RIGHT);
}

static void create_texture(int which) 
{
  static float texture_mapping_environment[] = {TV_DECAL, TV_NULL};
  static float prm[6][8] = {
    {
      TX_MAGFILTER,	TX_POINT, 
      TX_MINFILTER,	TX_POINT, 
      TX_WRAP,	TX_CLAMP, 
      TX_FAST_DEFINE, 
      TX_NULL
      }, 
    {
      TX_MAGFILTER,	TX_POINT, 
      TX_MINFILTER,	TX_BILINEAR, 
      TX_WRAP,	TX_REPEAT, 
      TX_FAST_DEFINE, 
      TX_NULL
      }, 
    {
      TX_MAGFILTER,	TX_POINT, 
      TX_MINFILTER,	TX_BILINEAR, 
      TX_WRAP,	TX_SELECT, 
      TX_FAST_DEFINE, 
      TX_NULL
      }, 
    {
      TX_MAGFILTER,	TX_BILINEAR, 
      TX_MINFILTER,	TX_BILINEAR, 
      TX_WRAP,	TX_CLAMP, 
      TX_FAST_DEFINE, 
      TX_NULL
      }, 
    {
      TX_MAGFILTER,	TX_BILINEAR, 
      TX_MINFILTER,	TX_BILINEAR, 
      TX_WRAP,	TX_REPEAT, 
      TX_FAST_DEFINE, 
      TX_NULL 
      }, 
    {
      TX_MAGFILTER,	TX_BILINEAR, 
      TX_MINFILTER,	TX_BILINEAR, 
      TX_WRAP,	TX_SELECT, 
      TX_FAST_DEFINE, 
      TX_NULL
      }
  };

  if(which == LEFT) 
    texdef2d(left_image, 3, TEXT_X_SIZE, TEXT_Y_SIZE, 
	     fleft_buffer, 0, prm[0]);
  else if (which == RIGHT) 
    texdef2d(right_image, 3, TEXT_X_SIZE, TEXT_Y_SIZE, 
	     fright_buffer, 0, prm[0]);

  tevdef(1000 + which, 0, texture_mapping_environment);
  tevbind(TV_ENV0, 1000 + which); 
}

void app_startup(void)
{
  int index;

  if(use_broker) {
    if(dojpegleft) {  
      index = find_gp(refl_name);
      if(index > -1) {
	printf("Calling add destination\n");
	add_dest(left_port, &gp_list[index]);
      } else  {
	printf("Couldn't find left reflector, index = %d\n", index);
      }
    } else {
      printf("Something claims we aren't doing jpegleft\n");
    }

    if(dojpegright) {
      index = find_gp(refr_name);
      if(index > -1) {
	add_dest(right_port, &gp_list[index]);
      } else {
	printf("Couldn't find right reflector, index = %d\n", index);
      }
    } else {
      printf("Something claims we aren't doing jpegright\n");
    }

    if(*master) {
      index = find_gp(robot_name);

      reset_robot(&gp_list[index], (*master)?idesk:arpa);
      *steering_turn = 90;

      if(index > -1)
	calibrate_joystick(joystick_bound, &switch_rad);
      
      index = find_gp(pt_name);
      if(index > -1)
	ptto(0, 0, &gp_list[index]);
      
      index = find_gp(svs_name);
      if(index > -1)
	set_stereo(&gp_list[index], in_stereo);
    }
  }
}

static void add_dest(int port, nexus_global_pointer_t *gp)
{
  nexus_buffer_t buffer;
  char itema = 'a', itemi = 'i', itemc = 'c';
  int count = 1, fa = port, len = 100, buffer_size;

  len = strlen(source_host);

  nexus_init_remote_service_request(&buffer, gp, "add-dest",
				    nexus_handler_hash("add-dest"));
  buffer_size = nexus_sizeof_char(&buffer, len);
  buffer_size += nexus_sizeof_int(&buffer, 4);
  buffer_size += nexus_sizeof_char(&buffer, 3);
  nexus_set_buffer_size(&buffer, buffer_size, 8);
  
  count = 2;
  nexus_put_char(&buffer, &itema, 1);
  nexus_put_int(&buffer, &count, 1);
  nexus_put_char(&buffer, &itemc, 1);
  nexus_put_int(&buffer, &len, 1);
  nexus_put_char(&buffer, source_host, len);
  nexus_put_char(&buffer, &itemi, 1);
  count = 1;
  nexus_put_int(&buffer, &count, 1);
  nexus_put_int(&buffer, &fa, 1);
  nexus_send_remote_service_request(&buffer);
}

static void calibrate_joystick(float *bound, float *switch_rad)
{
  int i;
  float maxx = 0.0, maxy = 0.0;

  for(i = 0; i < 500; i++) {
    maxx = MAX(maxx, CAVE_JOYSTICK_X);
    maxy = MAX(maxy, CAVE_JOYSTICK_Y);
  }
  
  bound[0] = maxx;
  bound[1] = maxy;

  bound[0] = 0.25;
  bound[1] = 0.25;
  *switch_rad = sqrt(pow(maxx, 2) + pow(maxy, 2));

#ifdef DEBUG_CALIB
  printf("X: %f Y: %f SR: %f\n", maxx, maxy, switch_rad);
#endif
}

/* Application display and supporting routines */
void app_display()
{
  /* Clear the last stuff in the buffer */
  czclear(0xff000000, getgdesc(GD_ZMAX));

  show_joystick(CAVE_JOYSTICK_X, CAVE_JOYSTICK_Y, switch_rad);

  show_menu();

  /* Build the texture we want on the wall */
  if(*in_stereo) {
    if(CAVEEye == CAVE_LEFT_EYE) {
      if(dojpegleft)
	build_texture(jleft, left_image);
    } else {
      if(dojpegright)     
	build_texture(jright, right_image);
    }    
  } else {
    if(dojpegleft)
      build_texture(jleft, left_image);
    if(dojpegright)     
      build_texture(jright, right_image);
  }

  if(*in_stereo) {
    translate(0.0, 0.0, -5.0);
    if(CAVEEye == CAVE_LEFT_EYE) {
      if (dojpegleft) {
	if(!(jleft->lock)) {
	  jleft->lock = 1;
	  texbind(TX_TEXTURE_0, left_image);
	  build_wall();
	  texbind(0, 0);
	  jleft->lock = 0;
	}
      }
    } else {
      if (dojpegright) {
	if(!(jright->lock)) {
	  jright->lock = 1;
	  texbind(TX_TEXTURE_0, right_image);
	  build_wall();
	  texbind(0, 0);
	  jright->lock = 0;
	}
      }
    }
  } else {
    pushmatrix();
    {
      translate(0.0, 0.0, -5.0);
      if (dojpegleft) {
	if(!(jleft->lock)) {
	  jleft->lock = 1;
	  texbind(TX_TEXTURE_0, left_image);
	  build_wall();
	  texbind(0, 0);
	  jleft->lock = 0;
	}
      }
    }
    popmatrix();
    pushmatrix();
    {
      scale(0.2, 0.2, 0.0);
      translate(-12.0, 4.0, 0.0);
      if (dojpegright) {
	if(!(jright->lock)) {
	  jright->lock = 1;
	  texbind(TX_TEXTURE_0, right_image);
	  build_wall();
	  texbind(0, 0);
	  jright->lock = 0;
	}
      }     
    }
    popmatrix();
  }
}

static void show_menu(void)
{
  char fdbck1[100];
  char ulist [50];

  /* User list */
  sprintf(ulist, "%s\n", idesk);
  if((*master && *control) || (*slave && !*control))
    cpack(0xff18f2ff);
  if((*master && !*control) || (*slave && *control))
    cpack(0xff7e00ff);
  cmov(2.5, 9.0, -4.0);
  charstr(ulist);
  sprintf(ulist, "%s\n", arpa);
  if((*master && *control) || (*slave && !*control))
    cpack(0xff7e00ff);
  if((*master && !*control) || (*slave && *control))
    cpack(0xff18f2ff);
  cmov(2.5, 8.4, -4.0);
  charstr(ulist);

  cpack(0xff0000ff);
  cmov(-5.0, 9.0, -4.0);
  charstr("Mode: ");
  sprintf(fdbck1, "%s", (*in_stereo)?"Stereo":"Mono");
  cpack(0xff00ffff);
  cmov(-3.0, 9.0, -4.0);
  charstr(fdbck1);
  
  switch(*activity) {
  case 0:
    sprintf(fdbck1, "Navigation");
    cpack(0xff00ffff);
    cmov(-3.0, 8.5, -4.0);
    charstr(fdbck1);
    break;
  case 1:
    sprintf(fdbck1, "Zoom");
    cpack(0xff00ffff);
    cmov(-3.0, 8.5, -4.0);
    charstr(fdbck1);
    break;
  case 2:
    sprintf(fdbck1, "Focus");
    cpack(0xff00ffff);
    cmov(-3.0, 8.5, -4.0);
    charstr(fdbck1);
    break;
  }

  if(*menuing > -1) {
    cpack(0xff007dff);
    cmov(-3.0, 8.0, -4.0);
    charstr(opts[*menuing]);
  }
}

static void show_joystick(float x, float y, float rad)
{
  float center[3] = { 3.0, 2.0, 0.0};
  float left[3] = { 2.4, 2.0, 0.0};
  float right[3] = { 3.6, 2.0, 0.0};
  float top[3] = { 3.0, 2.6, 0.0};
  float bottom[3] = { 3.0, 1.4, 0.0};
  float edge[3], wheelfront[3], wheelback[3];

  edge[0] = center[0] + x; edge[1] = center[1] + y; edge[2] = 0.0;
  wheelfront[0] = center[0] + (cosf(RAD(*steering_turn)) * 0.5);
  wheelfront[1] = center[1] + (sinf(RAD(*steering_turn)) * 0.5);
  wheelfront[2] = 0.0;

  wheelback[0] = center[0] - (cosf(RAD(*steering_turn)) * 0.5);
  wheelback[1] = center[1] - (sinf(RAD(*steering_turn)) * 0.5);
  wheelback[2] = 0.0;

  pushmatrix();
  {
    translate(0.0, 0.0, -3.5);

    bgnline();
    cpack(0xff0000ff);
    v3f(left);
    v3f(right);
    endline();
    
    bgnline();
    cpack(0xff0000ff);
    v3f(top);
    v3f(bottom);
    endline();

    /* Joystick position */
    bgnline();
    cpack(0x00ffff);
    v3f(center);
    v3f(edge);
    endline();


    /* Robot wheel direction */
    bgnline();
    cpack(0x00ff0000);
    v3f(wheelfront);
    v3f(wheelback);
    endline();

    cpack(0xff0000ff);
    circ(center[0], center[1], 0.5);
    
    cpack(0xffffff00);
    circ(center[0], center[1], rad);
  }
  popmatrix();
}

static void build_texture(myim *image, unsigned int textnum)
{
  int tsize;
  
  tsize = 320 * 256;

  /* register reader */
  image->lock = 1;
  if(image->getting_frames) {
    texbind(TX_TEXTURE_0, textnum);
    subtexload(TX_TEXTURE_0, textnum, 0.0, TVAL, 0.0, SVAL, tsize, 
	       (unsigned long *)image->data, 0);
    texbind(0, 0);
  }
  image->getting_frames = 0;
  image->lock = 0;
}

static void build_wall(void)
{
  static float normal[3] = {  0.0,  0.0, -1.0 };

  bgnpolygon();
  {
    /* Bottom Left */
    n3f(normal);
    t2f(texcoords[0]);
    v2f(coords[0]);
    
    /* Bottom Right */
    n3f(normal);
    t2f(texcoords[1]);
    v2f(coords[3]);

    /* Top Right */
    n3f(normal);
    t2f(texcoords[2]);
    v2f(coords[2]);
    
    /* Top Left */
    n3f(normal);
    t2f(texcoords[3]);
    v2f(coords[1]);
  }
  endpolygon();

#ifdef DEBUG_POLYGONS
  printf("Setting polygon at:\n(%f %f) (%f %f)\n(%f %f) (%f %f)\n",
	 coords[0][0], coords[0][1],
	 coords[1][0], coords[1][1],
	 coords[2][0], coords[2][1],
	 coords[3][0], coords[3][1]);
  printf("Setting texture at:\n(%f %f) (%f %f)\n(%f %f) (%f %f)\n",
	 texcoords[0][0], texcoords[0][1],
	 texcoords[3][0], texcoords[3][1],
	 texcoords[2][0], texcoords[2][1],
	 texcoords[1][0], texcoords[1][1]);
#endif
}

/* Application update and supporting routines */
void app_update()
{
  float azi, elev, roll;
  int index, other_index;

  if(*master)
    other_index = find_gp(arpa);
  else
    other_index = find_gp(idesk);
  
  nexus_poll();

  if(*control || *master) {
    if(CAVEButtonChange(2) == -1) {
      *menuing = 0;
      do_menu(switch_rad);
    }
  }

  if(*control) {
    if(*in_stereo) {
      if(*activity != 0) {
	*activity = 0;
	if(other_index > -1)
	  set_activity(&gp_list[other_index], activity);
	else
	  printf("Couldn't set other process activity navigation\n");
      }
      if(CAVEButtonChange(1) == -1) {
	CAVEGetWandOrientation(azi, elev, roll);
	index = find_gp(pt_name);
	if(index > -1)
	  update_pan_tilt(azi, elev, &gp_list[index]);
      }
      if(CAVEButtonChange(3) == -1) {
	index = find_gp(robot_name);
	if(index > -1)
	  beep_robot(&gp_list[index], (*master)?idesk:arpa);
      }
      //      index = find_gp(robot_name);
      //      if(index > -1)
      //	update_robot(CAVE_JOYSTICK_X, CAVE_JOYSTICK_Y, 
      //		     joystick_bound[0], joystick_bound[1], 
      //		     &gp_list[index], (*master)?idesk:arpa);
    } else {
      CAVEGetWandOrientation(azi, elev, roll);
      if(CAVEButtonChange(1) == -1) {
	index = find_gp(mpc_name);
	if(index > -1) {
	  *activity = 1;
	  if(other_index > -1)
	    set_activity(&gp_list[other_index], activity);
	  else
	    printf("Couldn't set other process activity to zoom\n");
	  if(elev > 0.0)
	    zoom_out(&gp_list[index]);
	  else
	    zoom_in(&gp_list[index]);	    
	}
      }
      if(CAVEButtonChange(3) == -1) {
	index = find_gp(mpc_name);
	if(index > -1) {
	  *activity = 2;
	  if(other_index > -1)
	    set_activity(&gp_list[other_index], activity);
	  else
	    printf("Couldn't set other process activity focus\n");
	  if(elev > 0.0)
	    focus_near(&gp_list[index]);
	  else
	    focus_far(&gp_list[index]);	    
	}
      }
    }
  }
}

static int do_menu(float switch_rad)
{
  int index, other_index;
  float dist;

  if(*master)
    other_index = find_gp(arpa);
  else
    other_index = find_gp(idesk);
  
  while(*menuing > -1) {
    if(CAVEButtonChange(2) == -1) {
      (*menuing)++;
      (*menuing) %= 7;
    }
    dist = sqrt(pow(CAVE_JOYSTICK_X, 2) + pow(CAVE_JOYSTICK_Y, 2));
    if(dist  > switch_rad) {
      switch(*menuing) {
      case 0:
	*in_stereo = 1;
	index = find_gp(svs_name);
	if(index > -1)
	  set_stereo(&gp_list[index], in_stereo);
	if(other_index > -1) {
#ifdef DEBUG_NEXUS
	  printf("About to set stereo to %s, %d\n",
		 nexus_global_pointer_string(&gp_list[other_index]), *control);
#endif
	  set_mode(&gp_list[other_index], in_stereo);
	} else {
	  printf("Couldn't find the other process to send stereo mode\n");
	}
	*menuing = -1;
	break;
      case 1:
	*in_stereo = 0;
	index = find_gp(svs_name);
	if(index > -1)
	  set_mono(&gp_list[index], in_stereo);
	if(other_index > -1) {
#ifdef DEBUG_NEXUS
	  printf("About to set mono to %s, %d\n",
		 nexus_global_pointer_string(&gp_list[other_index]), *control);
#endif
	  set_mode(&gp_list[other_index], in_stereo);
	} else {
	  printf("Couldn't find the other process to send mono mode\n");
	}
	*menuing = -1;
	break;
      case 2:
	/* Turn Left */
	index = find_gp(robot_name);
	if(*steering_turn < 180)
	  *steering_turn += 10;
	turn_robot(*steering_turn, &gp_list[index], (*master)?idesk:arpa);
	*menuing = -1;
	break;
      case 3:
	/* Turn Right */
	index = find_gp(robot_name);
	if(*steering_turn > 0)
	  *steering_turn -= 10;
	turn_robot(*steering_turn, &gp_list[index], (*master)?idesk:arpa);
	*menuing = -1;
	break;
      case 4:
	/* Move Forward */
	index = find_gp(robot_name);
	move_robot(6, 0, &gp_list[index], (*master)?idesk:arpa);
	*menuing = -1;
	break;
      case 5:
	/* Move Backward */
	index = find_gp(robot_name);
	move_robot(6, 1, &gp_list[index], (*master)?idesk:arpa);
	*menuing = -1;
	break;
      case 6:
	if(*control)
	  *control = 0;
	else
	  *control = 1;
	if(other_index > -1) {
#ifdef DEBUG_NEXUS
	  printf("About to set control to %s, %d\n",
		 nexus_global_pointer_string(&gp_list[other_index]), *control);
#endif
	  set_control(&gp_list[other_index], !*control);
	} else {
	  printf("Couldn't find the other process to send control\n");
	}
	*menuing = -1;
	break;
      default:
	*menuing = -1;
	break;
      }
    }
  }
  return(0);
}  
  
void app_exit(void)
{
  int i, other_index;

  other_index = find_gp((*master)?arpa:idesk);

  if(use_broker)
    for(i = 0; i < NUMNAMES; i++)
      nexus_destroy_global_pointer(&gp_list[i]);

  printf("Jleft %d\n", jleft->proc_id);

  if(jleft->proc_id) {
    kill(jleft->proc_id, SIGINT);
    kill(jleft->proc_id, SIGHUP);
    kill(jleft->proc_id, SIGKILL);
#ifdef DEBUG_PROCS
    printf("Killed Jpeg Receiver #1: %d\n", jleft->proc_id);
#endif
  }

  if(jright->proc_id) {
    kill(jright->proc_id, SIGINT);
    kill(jright->proc_id, SIGHUP);
    kill(jright->proc_id, SIGKILL);
#ifdef DEBUG_PROCS
    printf("Killed Jpeg Receiver #2: %d\n", jright->proc_id);
#endif
  }

  /* detach from the broker before committing suicide */
  if(use_broker) {
    unregister_module(&attach_gp, (*master)?idesk:arpa);
    nexus_destroy_global_pointer(&attach_gp);
  }
}

/* Utility functions ... */

static int fault_handler(void *user_arg, int fault_code)
{
  int i;
  
#ifdef DEBUG_NEXUS
	printf("Calling fault handler, butthead code: %d\n", fault_code);
#endif
  for(i =0; i < NUMNAMES; i++) {
    if(!nexus_is_null_global_pointer(&gp_list[i]))
      if(nexus_test_global_pointer(&gp_list[i])) {
#ifdef DEBUG_NEXUS
	printf("Destroying %s\n", nexus_global_pointer_string(&gp_list[i]));
#endif
	nexus_destroy_global_pointer(&gp_list[i]);
      }
  }
    return(0);
}

static int find_gp(char *name)
{
  int i;

  for(i = 0; i < NUMNAMES; i++) {
    if(!strcmp(name, name_list[i])) {
      return(i);
    }
  }

  return(-1);
}

/* Nexus required function, it registers our handlers */
int NexusBoot()
{
  nexus_register_handlers(handlers);
  return 0;
}

